/*
 * Copyright (c) 2022-2023 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                         Gianluca Corsini on Fri Sep 02 2022
 */
#include "acdxsim.h"

#include <err.h>
#include <math.h>
#include <pthread.h>

#include <iostream>
#include <string>
#include <vector>

#include "boost/bind.hpp"
#include "gazebo/gazebo.hh"
#include "gazebo/physics/physics.hh"
#include "gazebo/common/common.hh"

#include "libdynamixel.h"

using namespace gazebo;

/* compat macro to access renamed getters in gazebo >= 8 */
#if GAZEBO_MAJOR_VERSION < 8
# define GZGET(x)	Get ## x
#else
# define GZGET(x)	x
#endif

#define DXSIM_DEFAULT_PTY		"/tmp/pty-dynamixel"
#define DXSIM_DEFAULT_MOTOR_MODEL	"MX64T"
#define DXSIM_SUPPORTED_MOTOR_MODELS	"'MX64T', 'MX28T'"
#define DXSIM_MAXTORQUE			123456789.0


class dxsim_plugin : public ModelPlugin
{
  struct motor {
    physics::JointPtr joint;
    int spin;
    struct dynamixel_s *dynamixel;
    std::string model_name, eeprom;
    enum dx_model model_id;
    uint8_t id;
    enum dx_mode mode;
    double noise;
    bool servo;
    struct dynamixelstate_motor_s current, *desired;
  };

  struct log {
    bool now;
    uint8_t count, period;
  } log;

  std::vector<struct motor> motors;
  struct dynamixel_pty_s *pty;
  bool alive;

  pthread_t reader, updater;
  int32_t update_period;
  pthread_mutex_t dxsim_mutex = PTHREAD_MUTEX_INITIALIZER;

  event::ConnectionPtr begin, end;

  /* ------------ Methods ------------ */
  public:

  /* --- Load --- */
  void
  Load(physics::ModelPtr model, sdf::ElementPtr sdf)
  {
    std::string j_name, pty_name, spin;
    sdf::ElementPtr m, j, s;
    double p0, v0;
    motor motor;
    size_t i;

    /* pty */
    if (sdf->HasElement("pty"))
      pty_name = sdf->GetElement("pty")->GetValue()->GetAsString();
    else
      pty_name = DXSIM_DEFAULT_PTY;

    pty = dynamixel_init_pty();
    if (!pty) {
      warn("dxsim-gazebo cannot initialize pty");
      gzthrow("dxsim-gazebo cannot initialize pty");
    }
    if (dynamixel_open_pty(pty, pty_name.c_str())){
      warn("dxsim-gazebo cannot open pty");
      gzthrow("dxsim-gazebo cannot open pty");
    }
    warnx("dxsim-gazebo created pty `%s'", pty_name.c_str());

    /* motors */
    if (!sdf->HasElement("motors"))
      gzthrow("dxsim-gazebo missing <motors> tag");

    m = sdf->GetElement("motors");
    m = m->HasElement("motor") ? m->GetElement("motor") : NULL;
    for(; m; m = m->GetNextElement("motor")) {
      /* joint */
      if (!m->HasElement("joint")) {
        warnx("dxsim-gazebo ignoring <motor> without <joint> tag");
        continue;
      }
      j = m->GetElement("joint");
      j_name = j->GetValue()->GetAsString();
      motor.joint = model->GetJoint(j_name);
      if (!motor.joint) {
        warnx("dxsim-gazebo ignoring non-existent joint '%s'",
              j_name.c_str());
        continue;
      }
      if (motor.joint->GetType() !=
          (physics::Joint::JOINT | physics::Joint::HINGE_JOINT)) {
        warnx("dxsim-gazebo ignoring joint '%s' not of the hinge type",
              j_name.c_str());
        continue;
      }
      /* joint attribute spin */
      if (j->HasAttribute("spin")) {
        spin = j->GetAttribute("spin")->GetAsString();
        if (spin == "cw") motor.spin = -1;
        else if (spin == "ccw") motor.spin = 1;
        else {
          warnx("dxsim-gazebo ignoring unknown spin '%s' (must be cw or ccw)",
                spin.c_str());
          motor.spin = 1;
        }
      } else
        motor.spin = 1;

      /* initial condition */
      p0 = m->HasElement("p0") ? m->Get<double>("p0") : 0.;
      v0 = m->HasElement("v0") ? m->Get<double>("v0") : 0.;

#if GAZEBO_MAJOR_VERSION < 4
      motor.joint->SetAngle(0, p0);
#else
      motor.joint->SetPosition(0, p0);
#endif
      /* The 'SetVelocity' method does not guarantee to achieve the desired
       * speed in one iteration. To do so, in [1], it is suggested to use
       * `SetParam` of the parameters `fmax` and `vel`.
       * [1] https://classic.gazebosim.org/tutorials?tut=set_velocity
       */
      motor.joint->SetParam("fmax", 0, (double) DXSIM_MAXTORQUE);
      motor.joint->SetParam("vel", 0, v0);

      /* model */
      if (m->HasElement("model"))
        motor.model_name = m->GetElement("model")->GetValue()->GetAsString();
      else {
        warnx("motor model not specified, using default `%s'",
              DXSIM_DEFAULT_MOTOR_MODEL);
        motor.model_name = DXSIM_DEFAULT_MOTOR_MODEL;
      }
      if (motor.model_name == "MX28T")
        motor.model_id = DX_MX28T;
      else if (motor.model_name == "MX64T")
        motor.model_id = DX_MX64T;
      else {
        warnx("dxsim-gazebo ignoring unsupported motor model '%s'",
              motor.model_name.c_str());
        warnx("supported models are %s", DXSIM_SUPPORTED_MOTOR_MODELS);
        continue;
      }

      /* eeprom */
      if (m->HasElement("eeprom")) {
        motor.eeprom = m->GetElement("eeprom")->GetValue()->GetAsString();
      } else {
        motor.eeprom = DX_DEFAULT_EEPROM;
      }

      /* id */
      if (m->HasElement("id"))
        motor.id = (uint8_t) m->Get<int>("id");
      else
        motor.id = DX_DEFAULT_ID;
      if (motor.id == DX_BROADCAST_ID) {
        warnx("dxsim-gazebo invalid motor id '%u'", motor.id);
        continue;
      }

      /* noise */
      motor.noise =
        m->HasElement("noise") ? fabs(m->Get<double>("noise")) : 0.;
      /* servo */
      motor.servo = false;

      /* init dynamixel */
      motor.dynamixel = dynamixel_init(
        motor.model_id, motor.id, motor.eeprom.c_str());
      if (!motor.dynamixel) {
        warnx("dxsim-gazebo could not initialize motor id %u", motor.id);
        continue;
      }

      /* reload id and eeprom (modified during init) */
      motor.id = dynamixel_get_id(motor.dynamixel);
      motor.eeprom = dynamixel_get_eeprom_path(motor.dynamixel);
      motor.mode = dynamixel_get_op_mode(motor.dynamixel);

      /* init current and desired motor data */
      motor.current = {
        .effort   = 0.,
        .velocity = v0,
        .position = p0
      };
      motor.desired = dynamixel_get_motor_ref(motor.dynamixel);

      motor.servo = true;
      motors.push_back(motor);
    }

    /* create reader and updater thread */
    alive = true;
    update_period =
      1000000 * (sdf->HasElement("period") ? sdf->Get<double>("period") : 1e-3);
    pthread_create(&updater, NULL, updater_thread, this);
    pthread_create(&reader, NULL, reader_thread, this);

    /* plug events */
    begin = event::Events::ConnectWorldUpdateBegin(
      boost::bind(&dxsim_plugin::update_begin, this, _1));
    end = event::Events::ConnectWorldUpdateEnd(
      boost::bind(&dxsim_plugin::update_end, this));

    for(i = 0; i < motors.size(); i++)
      warnx("configured motor %10.10s, id %3u, eeprom %s",
            motors[i].model_name.c_str(),
            motors[i].id,
            motors[i].eeprom.c_str());
  }

  /* send data to Gazebo: update world state */
  void
  update_begin(const common::UpdateInfo &)
  {
    double noise;

    pthread_mutex_lock(&dxsim_mutex);
    for (uint8_t i = 0; i < motors.size(); i++) {
      if (motors[i].servo) {
        if (motors[i].noise > 0.0) {
#if GAZEBO_MAJOR_VERSION < 8
          noise = math::Rand::GetDblNormal(0., motors[i].noise);
#else
          noise = ignition::math::Rand::DblNormal(0., motors[i].noise);
#endif
        } else
          noise = 0.;

        switch (motors[i].mode) {
          case DX_POSITION_MODE:
          case DX_EXT_POSITION_MODE:
            motors[i].joint->SetPosition(
              0, motors[i].spin * (motors[i].desired->position + noise), true);
            break;

          case DX_VELOCITY_MODE:
            /* The 'SetVelocity' method does not guarantee to achieve the
             * desired speed in one iteration. To do so, in [1], it is
             * suggested to use `SetParam` of the parameters `fmax` and `vel`.
             * [1] https://classic.gazebosim.org/tutorials?tut=set_velocity */
            motors[i].joint->SetParam("fmax", 0, (double) DXSIM_MAXTORQUE);
            motors[i].joint->SetParam(
              "vel", 0, motors[i].spin * (motors[i].desired->velocity + noise));
            break;

          case DX_TORQUE_MODE:
            break;

          case DX_TORQUE_POSITION_MODE:
            break;

          default:
            warnx("undefined operating mode %u", motors[i].mode);
            break;
        }
      }
    }
    pthread_mutex_unlock(&dxsim_mutex);
  }

  /* retrieve data from Gazebo: update dynamixel state */
  void
  update_end()
  {
    size_t i;

    pthread_mutex_lock(&dxsim_mutex);
    for (i = 0; i < motors.size(); i++) {
      //motors[i].current.effort = joint->GetForceTorque(??);
      motors[i].current.velocity = motors[i].joint->GetVelocity(0);
      motors[i].current.position = motors[i].joint->Position(0);
    }
    pthread_mutex_unlock(&dxsim_mutex);
  }

  ~dxsim_plugin()
  {
    size_t i;

    pthread_mutex_lock(&dxsim_mutex);
    alive = false;
    for (i = 0; motors.size(); i++)
      dynamixel_fini(motors[i].dynamixel);
    pthread_mutex_unlock(&dxsim_mutex);
    pthread_join(reader, NULL);
    pthread_join(updater, NULL);
    dynamixel_fini_pty(pty);
  }


 private:

  static void *
  reader_thread(void *data)
  {
    dxsim_plugin *plugin = (dxsim_plugin *)data;
    motor *motor;
    size_t i;
    int s;

    pthread_mutex_lock(&plugin->dxsim_mutex);
    while(plugin->alive) {
      pthread_mutex_unlock(&plugin->dxsim_mutex);

      s = dynamixel_wait_msg(plugin->pty, 500);

      pthread_mutex_lock(&plugin->dxsim_mutex);
      if (!plugin->alive) break;

      if (s == 0) continue; /* timeout */
      else if (s < 0) {
        warnx("error when reading %s", dynamixel_pty_name(plugin->pty));
        continue;
      }

      while((s = dynamixel_get_msg(plugin->pty)) == 1) {
        for (i = 0; i < plugin->motors.size(); i++) {
          motor = &plugin->motors[i];
          if (motor->dynamixel)
            s = dynamixel_process_msg(motor->dynamixel, plugin->pty);
          else
            warnx("dxsim-gazebo could not access dynamixel id %u", motor->id);
        }
      }
      if (s < 0)
        warnx("cannot read message from %s", dynamixel_pty_name(plugin->pty));
    }
    pthread_mutex_unlock(&plugin->dxsim_mutex);

    warnx("dxsim disconnecting");
    return NULL;
  }

  static void *
  updater_thread(void *data)
  {
    dxsim_plugin *plugin = (dxsim_plugin *)data;
    struct timeval dl, tv;
    struct timespec req, now;
    motor *motor;
    size_t i;

    gettimeofday(&dl, NULL);

    pthread_mutex_lock(&plugin->dxsim_mutex);
    while(plugin->alive) {
      pthread_mutex_unlock(&plugin->dxsim_mutex);

      /* next deadline - skip missed periods */
      gettimeofday(&tv, NULL);
      do {
        dl.tv_usec += plugin->update_period;
        while (dl.tv_usec >= 1000000) {
          dl.tv_usec -= 1000000;
          dl.tv_sec++;
        }
      } while(dl.tv_sec < tv.tv_sec ||
              (dl.tv_sec == tv.tv_sec && dl.tv_usec < tv.tv_usec));

      /* sleep time */
      if (dl.tv_usec < tv.tv_usec) {
        tv.tv_usec -= 1000000;
        tv.tv_sec++;
      }
      req.tv_sec = dl.tv_sec - tv.tv_sec;
      req.tv_nsec = 1000 * (dl.tv_usec - tv.tv_usec);
      while(nanosleep(&req, &now)) {
        if (errno != EINTR) { warn("nanosleep"); break; }
        req = now;
      }

      pthread_mutex_lock(&plugin->dxsim_mutex);
      if (!plugin->alive) break;

      /* update motor stuff */
      for (i = 0; i < plugin->motors.size(); i++) {
        motor = &plugin->motors[i];
        if (motor->dynamixel) {
          motor->mode = dynamixel_get_op_mode(motor->dynamixel);
          motor->desired = dynamixel_get_motor_ref(motor->dynamixel);
          dynamixel_set_motor_state(motor->dynamixel, &motor->current);
        } else
          warnx("dxsim-gazebo could not access dynamixel id %u", motor->id);
      }
    }
    pthread_mutex_unlock(&plugin->dxsim_mutex);

    return NULL;
  }
};

/* register this plugin with the simulator */
GZ_REGISTER_MODEL_PLUGIN(dxsim_plugin);
